/* Test version using lex&yacc files */




/* (C) Copyright 1997 Albert Ludwigs University Freiburg
 *     Institute of Computer Science
 *
 * All rights reserved. Use of this software is permitted for 
 * non-commercial research purposes, and it may be copied only 
 * for that use.  All copies must include this copyright message.
 * This software is made available AS IS, and neither the authors
 * nor the  Albert Ludwigs University Freiburg make any warranty
 * about the software or its performance. 
 */


/*
 * main routine for extended graphplan algorithm
 *
 * basic algorithm written feb/mar/apr 1997 by joerg hoffmann
 *
 * inspired by graphplan and jana koehler
 *
 * does the following:
 *
 *           -  reads in specified operator and fact files
 *           -  removes uninstantiable operators
 *           -  calculates all instantiations of remaining operators
 *           -  builds graph until non-exclusive goals are reached first time
 *           -  searches for a plan:
 *                while ( not found plan at time ) do
 *                  build next graph layer;
 *                  time++;
 *                end
 *           -  prints final information to stdout
 *
 * additional features:
 * 
 *       by frank rittinger:
 *
 *           -  saves graph structure into text file for display, if wanted  
 *
 *       by michael brenner:
 *
 *           - removing irrelevant facts and operators. written may/june 1997
 *             inspired by theoretical works by bernhard nebel:
 *             powerful heuristics to deal with unnecessary information
 *             which would expand the graph and make problem unsolvable
 */


#include "ipp.h" /* defines, data structures, fn prototypes, global variables */
#include "rifo.h" /* all definitions for removing irrelevants needed by the files:
		    fbackchain.c, sets.c, rifohash.c, rifoutils.c, rifoio.c */

/*
 * global variables
 */


/* technical variables */
char *act_filename;
int lineno = 1;
inst_effect_list help_inst_effect_list;/* used for del_ALL_quantifiing */
BOOLEAN minimize = TRUE;/* has to do with recursive op choice in fn search */

/* globally important variables */

/* domain constants */
op_list loaded_ops;/* loaded, uninstantiated operators */
operator_list operators;/* list of instantiated operators */
token_list initial_facts;/* the initial facts */
effect_list raw_goal_facts;/* not yet preprocessed goal facts */
token_list goal_facts;/* the goal facts */
fact_list orig_constant_list;/* to store all typed objects */
fact_list orig_initial_facts;/* stores initials as fact_list */
fact_list negated_preds = NULL;/* needed for NOT preprocessing */ 
fact_list types;/* fast access to all objects of one type */

/* have to do with completeness check */
BOOLEAN same_as_prev_flag = FALSE;/* is set when graph has levelled off */
int first_full_time;/* stores time when graph has levelled off */
int previous_Scount = -1;/* previous num hashes at first full time */
int current_Scount = 0;/* current num hashes at that time */

/* the actual graph: two arrays of hashtables */
hashtable *fact_table, *op_table;

/* option preset values */
int max_nodes = 256;/* max number of nodes at one level */
BOOLEAN do_subset = TRUE;/* flag: use subset check in memoizing ? */
int display_info = 1;/* level of run time info printed */
int do_heuristic = 0;/* edge ordering strategy : preset none */
BOOLEAN rem_inertia = TRUE;/* remove inertia ? */

/* memoizing stuff */
root_table_type root_table[MAX_PLAN];/* one root tables at each time step */
/* just for info */
int simple_hits = 0;/* just for info: */
int partial_hits = 0;
int subset_hits = 0;/* different memoize types */

long bytes_graph = 0;

/* for searching */
goal_array *goals_at;/* used in searching; stores current goals at each level*/
goal_array *ops_at;/* stores currently choosen ops at each time */
goal_array *critics_at;/* stores current critics at each level */
int *num_ops_at;/* stores num of choosen ops */
int *num_critics_at;/* stores num of critics */
int num_of_actions_tried = 0;/* info */


/* variables for removing irrelevants */
operator_list save_operators; /* store loaded operators when running RIFO */
token_list save_initial_facts; 
int rifo_display_info = 1;
int mindepth = 1, maxdepth = 20;
int setlistthres = 10;
int unionstrategy = -1;
int oplevel = 1, factlevel = 1;
rifo_hashtable_t rel_fct_table, rel_op_table;
operator_list save_operators;
token_list constant_tl = NULL;  /* to store all typed objects as: name_type */
long memused = 0;
long memmaxused = 0;
long nodes_visited = 0;
rifo_hashtable_t objects_table;
settype *primary_fact_set, *secondary_fact_set; /* first and second part of relevant initial facts */


print_rifo_options()
{
  printf("\nRIFO (Remove Irrelevant Facts and Operators) Options for IPP:\n");
  printf("-rr <int>    global rifo options:\n");
  printf("     0     : only run ipp (default)\n");
  printf("     1     : use ipp with rifo\n");
  printf("     2     : stop after runnng rifo\n");
  printf("-ri <int>    run time information:\n"); 
  printf("     0     : none\n");
  printf("     1     : statistics (default)\n");
  printf("     2     : remaining initial facts\n");
  printf("     3     : remaining ground operators\n");
  printf("     4     : deleted initial facts, deleted ground operators\n");
  printf("     5     : complete backchaining process, debug information\n");
  printf("-rn <int>    specify min depth in backchaining (default 1)\n");
  printf("-rx <int>    to specify max depth in backchaining (default 20)\n");
  printf("-rt <int>    to specify threshold for set list length (default 10)\n");
  printf("-rm <int>    union method as in the paper: (default is 3)\n");
  printf("     1     : use all considered initial facts\n");
  printf("     2     : use all (stored) minimal sets \n");
  printf("     3     : use all cardinality-minimal sets, \n");
  printf("     4     : pick one cardinality-minimal set\n");
  printf("-rl <int>    combined selection as described in the paper:\n");
  printf("             (default is 1)\n");
  printf("     1     : select relevant objects\n");
  printf("     2     : select relevant initial facts\n");
  printf("     3     : select ground ops\n");
  printf("-rh          gives this list\n");

  exit(0);
}


int main( int argc, char *argv[] )

{

  int exit_status = 0; /* indicates at end if plan was found, used for PURIFY */
  char ops_file_name[MAX_LENGTH] = "";/* <ops_file_name>.ops */
  char fct_file_name[MAX_LENGTH] = "";/* <fct_file_name>.fct */
  char path[MAX_LENGTH]  = "";/* common path for both files */
  char ops_file[MAX_LENGTH]  = "";/* resulting name for ops file */
  char fct_file[MAX_LENGTH]  = "";/* same for fct file */

  char save_name[MAX_LENGTH] = "graph";/* preset saving name */

  token_list inertia;/* stores inertia literal names */

  BOOLEAN write_graph = FALSE;/* write graph to output files ? */
  int min_time = 0;/* specified minimal planning time */
  BOOLEAN reached_goals;/* stores return value of build_graph */
  BOOLEAN found_plan = FALSE;/* stores return value of search_plan */

  char option, rifo_option;/* command line interpretation */

  float build_time, plan_time, total_time;/* store times */
  struct tms start, end;/* helpers */

  int method, level; /* commmand line parameters */
  struct tms rifo_start, rifo_end;
  float rifo_time;
  int rifo_on=0; /* 0 rifo not used (default), 1 ipp+rifo, 2 stop after rifo */

  /* no options entered: print usage */
  if (argc == 1) {
    printf("\nusage of ipp:\n");

    printf("\nOPTIONS   DESCRIPTIONS\n\n");
    printf("-o <str>    operator file name\n");
    printf("-f <str>    fact file name\n");
    printf("-p <str>    path for operator and fact file\n\n");
    printf("-m <num>    max number of nodes at one level( preset: 256 )");
    printf("\n-t <num>  minimal planning time( preset: 0 )\n\n");
    printf("-S          don't do partial and complete subset check in memoizing\n");
    printf("-I          don't remove inertia from planning problem\n\n");
    printf("-h <num>    heuristic ordering of add edges( preset: 0 )\n");
    printf("     0    : none\n");
    printf("     1    : ( sum of pre values / num preconds )\n");
    printf("     2    : ( product of ( sum of pre values ) )\n\n");
    printf("-W          write complete graph to text files after planning\n");
    printf("-n <str>    specify name for graph output files( preset: graph )\n");
    printf("\n-i <num>  run-time information level( preset: 1 )\n");
    printf("     0    : only cpu time and num of actions info\n");
    printf("     1    : plan time info and final plan\n");
    printf("     2    : info on memoizing hits\n");
    printf("     3    : display loaded operators and facts\n");
    printf("     4    : show the instantiable ops\n");
    printf("     5    : show all the possible instantiations of those\n");
    printf("     6    : say which inertia are removed\n");
    printf("\nRun ipp with RIFO (Remove Irrellevant Facts and Operators):\n");
    printf("-rr <int>   global rifo options:\n");
    printf("     0    : only run ipp (default)\n");
    printf("     1    : use ipp with rifo\n");
    printf("     2    : stop after runnng rifo\n");
    printf("...to see all other RIFO options, just call IPP with option -rh\n" );

    printf("\nNOTE: if only an operator file <name> is specified,\n");
    printf("      default value for fact file is also <name>\n");
    printf("      similar if only fact file is specified.\n");
    printf("      suffixes '.ops' and '.fct' are added automatically.\n\n");
    exit( 1 );
  }

  /* read and interpret command line arguments */
  while ( --argc && ++argv ) {
    if ( *argv[0] == '-' && ( (strlen(*argv) == 2) || ( (*argv)[1] == 'r') ) ) {
      /*each option has 1 letter, except for RIFO options: -r + one letter */
      option = *++argv[0];
      switch ( option ) {/* first check the flags */
      case 'S':
        do_subset = FALSE;
        break;
      case 'W':
        write_graph = TRUE;
        break;
      case 'I':
        rem_inertia = FALSE;
        break;
      case 'r':
	rifo_option = *++argv[0];
	switch( rifo_option ) { /* RIFO options */
	case 'h':
	  print_rifo_options();
	  break;
	default:
	  if ( --argc && ++argv ) { /* option specified */
	    switch( rifo_option ) {
	    case 'l':
	      sscanf( *argv,"%d",&level );
	      switch ( level ) {
	      case 3:
		factlevel = 2;
		oplevel = 2;
		break;
	      case 2:
		factlevel = 2;
		oplevel = 1;
		break;
	      case 1:
		factlevel = 1;
		oplevel = 1;
		break;
	      case 0:
		factlevel = 0;
		oplevel = 0;
		break;
	      }
	      break;
	    case 'u':
	      sscanf( *argv,"%d",&unionstrategy );
	      break;
	    case 'm':
	      sscanf( *argv,"%d",&method );
	      switch ( method ) {
	      case 1:
		unionstrategy = 0;
		break;
	      case 2:
		unionstrategy = -10000;
		break;
	      case 3:
		unionstrategy = -1;
		break;
	      case 4:
		unionstrategy = 1;
		break;
	      }
	      break;
	    case 'r':
	      sscanf( *argv,"%d",&rifo_on );
	      break;
	    case 'i':
	      sscanf( *argv,"%d",&rifo_display_info );
	      break;
	    case 'n':
	      sscanf( *argv,"%d",&mindepth );
	      break;
	    case 'x':
	      sscanf( *argv,"%d",&maxdepth );
	      break;
	    case 't':
	      sscanf( *argv,"%d",&setlistthres );
	      break;
	    }
	  } else {/* something wrong with the option */
	    printf("\nipp: wrong usage with rifo options\n\n");
	    exit( 1 );
	  }
	  break;
	} 
	break; /* end of RIFO options */
      default:
        if ( --argc && ++argv ) { /* option specified */
	  switch ( option ) { /* examine the option letter */
	  case 'o':
	    strncpy(ops_file_name, *argv, MAX_LENGTH);
	    break;
	  case 'f':
	    strncpy(fct_file_name, *argv, MAX_LENGTH);
	    break;
          case 't':
            sscanf(*argv, "%d", &min_time);
            break;
          case 'm':
            sscanf(*argv, "%d", &max_nodes);
            break;
          case 'i':
            sscanf(*argv, "%d", &display_info);
            break;
          case 'h':
            sscanf(*argv, "%d", &do_heuristic);
            break;
          case 'p':
	    strncpy(path, *argv, MAX_LENGTH);
            break;
          case 'n':
	    strncpy(save_name, *argv, MAX_LENGTH);
            break;
	  default:
	    printf( "\nipp: unknown option: %c entered\n\n", option );
	    exit( 1 );
	    break;
	  }
        } else {/* something wrong with the option */
	  printf("\nipp: unknown option\n\n");
	  exit( 1 );
        }
      }
    } else {/* option has more than one letter or no '-' prefix */
      printf("\nipp: wrong usage\n\n");
      exit( 1 );
    }
  }

  /* unknown heuristc entered */
  if (  do_heuristic > 2 ) {
    printf( "\nipp: unknown heuristic entered\n\n" );
    exit( 1 );
  }


 /* too large max_nodes as argument of -m specified (Jana 17/9) */ 
  if (  max_nodes > MAX_MAX_NODES ) {
    printf( "\nipp: -m %d exceeds MAX_MAX_NODES bound of %d\n", 
               max_nodes,  MAX_MAX_NODES );  
    exit( 1 );
  }

  /* no input name at all entered */
  if ( !ops_file_name[0] && !fct_file_name[0] ){
    printf( "\nipp: at least one inputfile name needed\n\n" );
    exit( 1 );
  }

  /* only one entered ? => copy from other */
  if ( !ops_file_name[0] ) strncpy( ops_file_name, fct_file_name, MAX_LENGTH);
  if ( !fct_file_name[0] ) strncpy( fct_file_name, ops_file_name, MAX_LENGTH);

  /* add type suffixes */
  strcat( ops_file_name, ".ops" );
  strcat( fct_file_name, ".fct" );

  /* add path info, complete file names will be stored in
   * ops_file and fct_file
   */
  strcat( ops_file, path );
  strcat( fct_file, path );
  strcat( ops_file, ops_file_name );
  strcat( fct_file, fct_file_name );

  load_fct_file( fct_file );
  load_ops_file( ops_file );
  
  /* preprocessing: enable to use negated facts */
  preprocess_not();  /* globals changed: negated_preds, initial_facts */

  /* convert initials to graph storage */
  initial_facts = token_list_from_fact_list( orig_initial_facts );

  /* preprocessing: remove quantifiers, remove
   eq() predicates */
  goal_facts = dequantify_goals( raw_goal_facts, &loaded_ops );

  /* display info on loaded operators and fact, if wanted */
  if ( display_info > 2 ) {
    printf( "\nloaded operators are:\n" );
    print_ops( loaded_ops );
    printf( "\nloaded facts are:\n" );
    print_fct( orig_constant_list, initial_facts, goal_facts );
    printf( "\n" );
  }

  /* remove operators that cannot be instantiated with given facts */
  remove_uninstantiable_ops( &loaded_ops, orig_constant_list );

  /* display info on not removed operators, if wanted */
  if ( display_info > 3 ) {
    printf( "\nnot removed operators are:\n" );
    print_ops( loaded_ops );
    printf( "\n" );
  }

  /* just for fun ( makes output look nicer ) */
  if ( display_info < 3 ) printf( "\n" );

  /* begin timing */
  times( &start );
  
  /* find out which literals are inertia */
  inertia = get_inertia( orig_initial_facts, loaded_ops );

  /* calculate all possible instantiations 
   * and store them in global list operators
   */
  instantiate( loaded_ops, orig_initial_facts, inertia, orig_constant_list );  

  /* find and remove all facts that are never deleted from the
   * global op and fact lists
   */
  if ( rem_inertia )
    handle_inertia( inertia );

  /* this one goes through all instantiated operators, looking
   * for effects with identical conditions that might occur due to
   * removal of inertia or eq - conditions
   *
   * is defined in inertia.c, which doesn't make sense but saves us from
   * yet another file of c code
   */
  merge_identical_effects(); 

  /* display info on instantiated operators, if wanted */
  if ( display_info > 4 ) {
    printf( "\n Original instantiated operators are:\n\n" );
    print_operators( operators );
    printf("\n" );
  }

  /* store all information that might be changed by rifo
     in order to rebuild it (using reset_original_ipp_information() )
     if rifo strategy fails */
  save_original_ipp_information();

  /* remove irrelevant facts and operators. 
     globals changed     : operators, initial_facts;
     restore old values with reset_original_ipp_information() */
  if ( rifo_on ) {
    reset_original_ipp_information();
    times( &rifo_start );
    find_relevant_initial_facts();
    times( &rifo_end );
    rifo_time = ( float ) 
      ( ( rifo_end.tms_utime - rifo_start.tms_utime +
	  rifo_end.tms_stime - rifo_start.tms_stime  ) / 100.0 ) ;
    if ( rifo_display_info )
      {
	printf( "rifo: %.2f secs used to remove irrelevant facts and operators\n", 
		rifo_time );
	printf( "      %.2f MBytes\n",( double )memused/( 1024.0*1024.0 ) );
	printf( "      %ld search nodes visited\n\n",nodes_visited );
      }
    if ( rifo_on == 2 ) /* just see rifo working and stop afterwards */
      exit( 0 );
  }
  
  /* build graph until non-exclusive goals are reached */
  reached_goals = build_graph( &min_time );

  /* get building time */
  times( &end );
  build_time = ( float ) ( ( end.tms_utime - start.tms_utime +
                             end.tms_stime - start.tms_stime  ) / 100.0 ) ;
  times( &start );

  if ( !reached_goals ) {
    if ( min_time < MAX_PLAN ) {
      /* graph has leveled off and goals are still not reached */
      printf("\nipp: problem unsolvable: can't reach non exclusive goals\n\n");
    } else {
      /* we didn't even have time to find the goals */
      printf( "\nipp: MAX_PLAN too small( preset value: %d )\n\n", MAX_PLAN );
    }
  } else {
    if ( min_time ) {
      /* found goals at level higher than 0; start searching for a plan */
      for ( ; min_time < MAX_PLAN; min_time++ ) {
        if ( ( found_plan = search_plan( min_time ) ) == TRUE ) break;
        /* couldn't find plan; check for unsolvability */
        if ( same_as_prev_flag ) {
          /* start after graph has leveled off */
          if ( previous_Scount == current_Scount ) {
            /* nothing has changed in last step -> unsolvable */
            break;
          } else {
            /* still improvements going on */
            previous_Scount = current_Scount;
          }
        }
        /* we must continue search with one more layer */
        build_graph_layer();
      }
      /* search finished; */
      if ( found_plan ) {
        /* either we found a plan, */
        if ( display_info ) print_plan( min_time );
	/* print info on memoizing hits, if wanted */
	if ( display_info > 1 ) {
	  printf("\nipp: had %7d simple memoizing hits", simple_hits);
	  if ( do_subset ) {
	    printf("\nipp  had %7d partial memoizing hits", partial_hits);
	    printf("\nipp: had %7d subset memoizing hits", subset_hits);
          }
	  printf("\n\n");
	}
      } else {
        if ( min_time < MAX_PLAN ) {
          /* or there is no plan for the problem, */
          printf( "\nipp: problem proved unsolvable\n\n" );
	  /* print info on memoizing hits, if wanted */
	  if ( display_info > 1 ) {
	    printf("\nipp: had %7d simple memoizing hits", simple_hits);
	    if ( do_subset ) {
	      printf("\nipp: had %7d partial memoizing hits", partial_hits);
	      printf("\nipp: had %7d subset memoizing hits", subset_hits);
            }
	    printf("\n\n");
	  }
        } else {
          /* or we didn't give the planner enough time */
          printf("\nipp: MAX_PLAN too small( preset value: %d )\n\n",MAX_PLAN);
        }
      }
    } else {/* corresponds to if ( min_time ) */
      printf( "\nipp: goals are contained in initial state\n\n" );
    }
  }

  /* calculate and print running time */
  times( &end );
  plan_time = ( float ) ( ( end.tms_utime - start.tms_utime +
                            end.tms_stime - start.tms_stime  ) / 100.0 ) ;

  printf( "\nipp: number of actions tried: %8d", num_of_actions_tried );

  printf( "\nipp: spent %.2f seconds initial building graph", build_time);
  printf( "\nipp: spent %.2f seconds searching for plan", plan_time);
  total_time = build_time + plan_time;
  printf( "\nipp: spent %.2f seconds total time\n", total_time);

  if ( display_info )
    printf("\nmemory use for graph: %.1f Kbytes\n\n",
	   (float) bytes_graph/1024);

  /* create the graph files, if required */
  if ( write_graph ) {
    printf("\nipp: writing graph\n\n");
    if ( !SaveGraph( save_name, min_time ) ) {
      printf("\nipp: error writing graph\n\n");
    }
  }

  if (found_plan)
    {
      exit_status = 0;
    }
  else
    {
      exit_status = 1;
    }

  exit( exit_status );

}
