/* (C) Copyright 1997 Albert Ludwigs University Freiburg
 *     Institute of Computer Science
 *
 * All rights reserved. Use of this software is permitted for 
 * non-commercial research purposes, and it may be copied only 
 * for that use.  All copies must include this copyright message.
 * This software is made available AS IS, and neither the authors
 * nor the  Albert Ludwigs University Freiburg make any warranty
 * about the software or its performance. 
 */


/*
 * fn for searching:
 *   BOOLEAN search_plan( int )
 *
 * makes use of the following functions:
 *   int setup( int ),
 *   void mark_exclusives( vertex_list ),
 *   void unmark_exclusives( vertex_list ),
 *   BOOLEAN is_deleted( vertex_list, int, vertex_list ),
 *   BOOLEAN goals_still_possible( int, int ),
 *   edge_list without_new_goals( edge_list ),
 *   BOOLEAN action_set_is_minimal( int ),
 *   index_list_of_edge_list make_index( list_of_edge_list, int )
 * taken from util_search.c,
 *   void memoize( int, goal_array, int, goal_array, int ),
 *   BOOLEAN memoized( int, goal_array, int, goal_array, int ),
 *   BOOLEAN partial_memoized( int, goal_array, int, goal_array, int ),
 *   BOOLEAN subset_memoized( int, goal_array, int, goal_array, int )
 * taken from memoize.c
 */

#include"ipp.h"/* defines, data structures, fn prototypes, global variables */

/* the main recursive search function, makes use of all the
 * functions that are declared here.
 *
 * see there for descriptions
 *
 * NOTE: what i call 'critics' in this code, is called 'negative goals'
 *       in jana's paper
 */
BOOLEAN search( int curr_index, int next_index, int time,
                int critic_index, list_of_edge_list SS, BOOLEAN prevent );

BOOLEAN choose_critics( int time, list_of_edge_list SS, BOOLEAN initialise );

list_of_edge_list compute_SS( int time, list_of_edge_list SS );

BOOLEAN try_op( vertex_list op, edge_list conditions,
                int curr_index, int *next_index, int time );

void untry_op( vertex_list op, edge_list conditions,
               int *next_index, int time );

BOOLEAN complete_critics_and_search( BOOLEAN mode, int time, int next_index,
                                     list_of_edge_list SS );


/* the setting up function; called by main
 * simply stes up search and does the initial call
 * as well as memoizing the result if search was a failure
 */
BOOLEAN search_plan( int max_time )

{

  int num_goals, i;
  BOOLEAN result;

  /* sets up some important things; see there */
  num_goals = setup( max_time );

  /* say that these are our goals */
  for ( i=0; i < num_goals; i++ ) goals_at[max_time][i]->is_goal = 1;
  /* say that we didn't choose any op( at time - 1 ) so far */
  num_ops_at[max_time-1] = 0;

  /* try to find a plan */
  result = search( num_goals - 1, 0, max_time, 0, NULL, FALSE );

  /* undo goal info */
  for ( i=0; i < num_goals; i++ ) goals_at[max_time][i]->is_goal = 0;
 
  /* if it was a failure, memoize that */
  if ( !result )  memoize( max_time,
                           goals_at[max_time], num_goals,
                           critics_at[max_time], 0 );

  /* and tell main what we found */
  return result;

}


/* this is a terribly complicated function, basically divided into two
 * parts:
 *   - first, choose ops that make all goals in
 *     goal_array goals_at[time] from 0 to curr_index true;
 *     this is done recursive: choose op for goals_at[time][curr_index],
 *     then call again with curr_index - 1( and next_index higher, see now )
 *     meanwhile, set up the new goals by putting the ( num ) facts in at
 *     goal_array[time-1][next_index] to goal_array[time-1][next_index+num].
 *   - after that is finished( curr_index == -1 ), go on to next
 *     time step: 
 *     - first we have to look if there is a noop adding a critic;
 *       if so, see if it is exclusive of an op that we use already.
 *       if there is no such op, try to prevent this critic one step
 *       earlier. if this was unsuccessfull, try out all ops that
 *       delete that critic. that's the sense of params critic_index,
 *       SS and prevent:
 *       critic_index says which critics we still have to care for,
 *       SS stores the so far propagation of critics to
 *       prevent one step earlier, prevent is to turn off minimality
 *       check when we are in the prevent - pattern.
 *     - now we finally go on to the next step:
 *       compute SS, choose all possible critics one after the other,
 *       ( complete them, see there ), try to find a plan with 
 *       new goals and critics one step earlier( if they are not
 *       memoized already ), if we fail for all possibilities: return FALSE
 */ 
BOOLEAN search( int curr_index, int next_index, int time,
                int critic_index, list_of_edge_list SS, BOOLEAN prevent )

{

  edge_list help;/* index and helper */
  cond_edge_list i_ce;/* index for current chosen add edge */
  int i;/* index */
  list_of_edge_list new_SS = SS, temp;/* for propagating critics */

  BOOLEAN found_critics;/* says if we found a critic set */

  vertex_list op, noop, ft;/* helpers */

  /* don't have to make goals at initial level true,
   * cause they are. we found a plan.
   */
  if ( time == 0 ) return TRUE;

  /* all ops for making goals true are chosen.
   * 
   * do some more computation and go on to next level.
   */
  if ( curr_index < 0 ) {

    /* the minimality check; see if there's an op that we can leave out
     * of the plan, i.e., that doesn't add a goal alone;
     *
     * NOTE: if the op is in the plan to delete a critic, then this is
     *       obviously nonsense. that's why we need param prevent, which
     *       is true when we are in the prevent recursion.
     */
    if ( !prevent && !action_set_is_minimal( time ) ) return FALSE;

    /* check all critics that aren't looked after already
     */
    for ( i=critic_index - 1; i>-1; i-- ) {
      ft = critics_at[time][i];
      /* only try preventing critics that aren't marked */
      if ( cant_do_ft( ft, time ) ) continue;
      noop = ft->my_noop;
      if ( !noop ) continue;
      if ( !cant_do_op( noop, time-1 ) ) {
        /* now we got a critic that is added by a noop that's
         * not exclusive of any op we have chosen so far;
         * first try to propagate the critic to the next time step;
         * this doesn't make sense if the critic is a new goal
         */
	if ( !(ft->prev_time->is_goal) ) {

	  /* add an edgelist consisting of our critic to SS
	   * and call up search with this new SS on remaining critics
	   */
	  help = ( edge_list ) calloc( 1, sizeof( edge_list_elt ) );
	  CHECK_MEMORY(help);
	  help->endpt = ft->prev_time;
	  help->next = NULL;
	  temp = ( list_of_edge_list )
	    calloc( 1, sizeof( list_of_edge_list_elt ) );
	  CHECK_MEMORY(temp);
	  temp->list = help;
	  temp->op = noop;
	  temp->next = new_SS;
	  new_SS = temp;
	  
	  /* now try to find a plan with extended SS and lower critic_index;
	   *
	   * don't need the minimality check anymore, we've done it already
	   * ( talking bout last argument in recursive call )
	   */
	  if ( search( -1, next_index, time, i, new_SS, TRUE ) )
	    return TRUE;
	  free(temp);  /* Jana/Frank to fix memory leak 17/9 */
	  free(help); /* Jana/Frank to fix memory leak */
	}

        /* now try all ops that delete our critic; first see
         * if an op that we use already deletes it ( has to be a conditional
         * delete then )
         */
        for ( i_ce = ft->del_edges; i_ce; i_ce = i_ce->next ) {
	  op = i_ce->endpt;
	  if ( !op->is_used ) continue;

          /* try to put it into the plan */
	  if ( !try_op( op, i_ce->conditions, curr_index, &next_index, time ) )
            continue;

          /* put it in: search with remaining critics,
           * if we found a plan then we have a valid plan here, too
           *
           * NOTE: SS can be non-empty, for we are in the
           *       put-in-prevent-op recursion pattern.
           */
	  if ( search( -1, next_index, time, i, SS, TRUE ) )
            return TRUE;

          /* search was a failure: remove that op and try the next one */
	  untry_op( op, i_ce->conditions, &next_index, time );
	}

        /* now do exactly the same thing with the ops we didn't use so far
         */
        for ( i_ce = ft->del_edges; i_ce; i_ce = i_ce->next ) {
	  op = i_ce->endpt;
	  if ( op->is_used ) continue;

	  if ( !try_op( op, i_ce->conditions, curr_index, &next_index, time ) )
            continue;

	  if ( search( -1, next_index, time, i, SS, TRUE ) )
            return TRUE;

	  untry_op( op, i_ce->conditions, &next_index, time );
	}

        /* all attempts to prevent this critic have failed
         */
        return FALSE;
      }
    }

    /* empty goal sets on non initial levels dont make sense
     */
    if ( next_index == 0 && time > 1) {
      printf("\nipp: empty goal set on non initial level encountered\n\n");
      exit( 1 );
     }

    /* now compute SS, adding condition lists to the edge of edge list
     * that we have already
     */
    SS = compute_SS( time, SS ); 

    /* no point in trying to set up critics at time step 0;
     * can't make initial facts untrue
     *
     * NOTE: if it is time 1 here, the critics we choose will be
     *       at time 0, cause SS contains condition lists of
     *       effects that make facts at time 1 true.
     */
    if ( time == 1 && !each_list_contains_dummy( SS ) ) return FALSE;
    
    /* this is the loop which tries all possible choices of critics;
     *
     * description of choose critics see there
     */
    for ( found_critics = choose_critics( time, SS, INIT );
          found_critics;
          found_critics = choose_critics( time, SS, EXEC) ) {

      /* now it's time for yet another recursion pattern:
       *   it can be that a chosen op adds a new critic of a
       *   different chosen op.( this is not allowed cause
       *   they ought to be carried out parallel )
       *   if so unconditionally, return FALSE,
       *   if so conditionally, we have to prevent that effect
       *   by putting a new critic into our critics; done that,
       *   we have to do the same check again->recursion.
       *
       * NOTE: the recursive call on search( time-1 ) is done in the
       *       inner recursion pattern of this function, namely
       *       when it's first argument is SEARCH.
       */
      if ( complete_critics_and_search( COMPLETE, time, next_index, SS ) )
        return TRUE;

    }

    /* tried all possibilities without success: return FALSE */
    return FALSE;
  }

  /* this is the beginning of the part which chooses ops for making
   * the goals true.
   */

  /* skip over any goals that have been made true just by luck
   * by any op we've chosen already: is_true value > 0
   */
  while ( ( ft = goals_at[time][curr_index] )->is_true ) {
    curr_index--;
    if ( curr_index < 0 ) {
      /* invoke the try-prevent-op-pattern with SS = NULL and
       * all critics to go.
       */
      return search( -1, next_index, time, critic_index, NULL, FALSE );
    }
  }

  /* try all possibilities of making that fact true */
  for ( i_ce = ft->add_edges; i_ce; i_ce = i_ce->next ) { 

    /* try to put op into plan */
    op = i_ce->endpt;
    if ( !try_op( op, i_ce->conditions, curr_index, &next_index, time) )
      continue;


    /* try making the remaining goals true */
    if ( search( curr_index - 1, next_index, time, critic_index, NULL, FALSE ) )
      return TRUE;
    
    /* didn't manage: put op out of plan and try next one */
    untry_op( op, i_ce->conditions, &next_index, time );

  }

  /* tried all add edges for fact without success: return FALSE */
  return FALSE;

}


/* this is a recursive algorithm for 'completing critics',
 * abstract description see where it's called in search.
 */
BOOLEAN complete_critics_and_search( BOOLEAN mode, int time, int next_index,
                                     list_of_edge_list SS )

{

  int i, j;
  edge_list i_e, temp;
  cond_edge_list i_ce;
  vertex_list op,ft;
  
  if ( mode == SEARCH ) {
 
    /* the completion is finished; continue search one step earlier now,
     * first check if goals and critics have been memoized already.
     * ( if wanted, do subset checks also, turned on by option -S )
     *
     * ( NOTE: it is always more difficult to solve problems with more
     *         goals / critics, so if we proved a subset unsolvable, we
     *         can back right up )
     */ 
    if ( next_index > 0 && memoized( time - 1,
			goals_at[time-1], next_index,
			critics_at[time-1], num_critics_at[time-1] ) ) {
      return FALSE;
    }

    /* this is technical for setting up the ops_at - array;
     * time - 2 cause we're gonna make goals at time - 1 true, so we
     * will choose ops at time - 2.
     */
    if ( time > 1 ) num_ops_at[time-2] = 0;
	
    /* now finally try to solve new goals and critics one step earlier
     *
     * NOTE:  that's the point in the search recursion where
     *        critic_index is initialised; it's value( the number
     *        of critics we have ) is maintained until the prevent
     *        recursion pattern comes up.
     */
    if( search( next_index - 1, 0, time - 1,
		num_critics_at[time-1], NULL, FALSE ) ) return TRUE;

    /* failed -> memoize this result */
    memoize( time-1,
	     goals_at[time-1], next_index,
	     critics_at[time-1], num_critics_at[time-1] );
    return FALSE;

  } else {

    /* we still have to complete the critics
     */

    /* no SS -> no critics -> no completion needed */
    if ( !SS )
      return complete_critics_and_search( SEARCH, time, next_index, SS );

    /* for all used ops do */
    for ( i=0; i<num_ops_at[time-1]; i++ ) {
      op = ops_at[time-1][i];
      /* for all add edges do */
      for ( i_ce = op->add_edges; i_ce; i_ce = i_ce->next ) {
	ft = i_ce->endpt->prev_time;
	if ( !ft || !ft->is_critic ) continue;
        /* the fact is a new critic */
	if ( !i_ce->conditions ) {
          /* we add it unconditionally */
	  for ( j=0; j<ft->is_critic; j++ )
            if ( (ft->is_critic_for)[j] != op ) break;
          if ( j<ft->is_critic ) {
            /* it was someone else's critic
             */
	    return FALSE;
	  }
	} else {
          /* it is a conditional effect */

          /* if we already prevent this effect, there's nothing to do */
	  for ( i_e = i_ce->conditions; i_e; i_e = i_e->next )
	    if ( i_e->endpt->is_critic ) break;
	  if ( i_e ) continue;

          /* check if it's someone else's critic */
	  for ( j=0; j<ft->is_critic; j++ )
            if ( (ft->is_critic_for)[j] != op ) break;
          if ( j<ft->is_critic ) {

            /* can only prevent facts that are no new goals */
	    temp = without_new_goals( i_ce->conditions );
	    if ( !temp ) {
              /* conditions contained in new goals: back right up */
	      return FALSE;
	    }

            /* put conditions into new critics one after the other */
	    for ( i_e = temp; i_e; i_e = i_e->next ) {
	      critics_at[time-1][num_critics_at[time-1]++] = i_e->endpt;
	      i_e->endpt->is_critic++;
	      (i_e->endpt->is_critic_for)[i_e->endpt->is_critic-1] = op;

              /* try to find a plan with these new critics, i.e.,
               * first finish completion pattern.
               */
	      if ( complete_critics_and_search( COMPLETE, time,
                                                next_index, SS ) )
                return TRUE;
              
              /* failed: put the critic out and try next one */
	      (i_e->endpt->is_critic_for)[i_e->endpt->is_critic--] = NULL;
	      critics_at[time-1][--num_critics_at[time-1]] = NULL;
	    }

            /* tried all possibilities without success: return FALSE */
	    return FALSE;
	  }
	}
      }
    }/* matches for all used ops do */

    /* we didn't have to add a new critic: we are finished with completing */
    return complete_critics_and_search( SEARCH, time, next_index, SS );

  }

}


/* this one checks if a new op can be put into the plan, i.e.
 * if it's not exclusive of any other chosen op etc.
 */
BOOLEAN try_op( vertex_list op, edge_list conditions,
        	int curr_index, int *next_index, int time )

{

  edge_list i_e, j_e;/* indizes */
  cond_edge_list i_ce;/* index */
  int was_used;/* to tell if a fact was already there */
  vertex_list ft_;/* helper */

  /* see if op is marked as being exclusive */
  if ( cant_do_op( op, time-1 ) ) return FALSE; 

  /* if we don't use this op yet */
  if ( !op->is_used ) {
    /* see if it unconditionally adds a critic */
    for ( i_ce = op->add_edges; i_ce; i_ce = i_ce->next )
      if ( !i_ce->conditions && i_ce->endpt->is_critic ) break;
    if ( i_ce ) return FALSE;

    /* see if it unconditionally deletes a goal or a new goal
     *
     * NOTE: the op in fact can be non-exclusive of the others
     *       and, in spite of that, delete the add effect of
     *       another one for we got conditional effects
     */
    for ( i_ce = op->del_edges; i_ce; i_ce = i_ce->next ) {
      if ( i_ce->conditions ) continue;
      ft_ = i_ce->endpt;
      if( ft_->is_goal ) break;
      if( ft_->prev_time && ft_->prev_time->is_goal ) break;
    }
    if ( i_ce ) return FALSE;

    if ( !goals_still_possible( curr_index, time, op ) ) {
      return FALSE;
    }
  }

  /* now check the preconds for exclusion of any other new goals
   * ( they can be exclusive of effect conditions we committed to )
   */
  for ( i_e = op->precond_edges; i_e; i_e = i_e->next ) {
    if ( cant_do_ft(i_e->endpt, time - 1 ) ) {
      return FALSE;
    }
  }

  /* do the same thing with the effect conditions( if we got some )
   * also check if they are unconditionally deleted by some
   * op we've already chosen, attention: by some op different from
   * the one we're currently working on.
   */
  for ( i_e = conditions; i_e; i_e = i_e->next ) {
    if ( cant_do_ft(i_e->endpt, time - 1 ) ||
	 is_deleted( i_e->endpt, time, op ) ||
	 i_e->endpt->is_dummy ) {
      return FALSE;
    }
  }

  /* now it's clear that we can put the op into the plan */

  /* mark the op and put it into the global array */
  was_used = op->is_used++;
  if ( !was_used && num_ops_at[time-1] == MAX_GOALS ) {
    printf( "\n\nipp: increase MAX_GOALS( preset value: %d )",MAX_GOALS );
    exit( 1 );
  }
  if( !was_used ) {
    ops_at[time-1][num_ops_at[time-1]++] = op;
    num_of_actions_tried++;
  }

  /* also put the preconditions into goal array at time - 1 */
  for ( i_e = op->precond_edges; i_e; i_e = i_e->next ) {
    was_used = i_e->endpt->is_goal++;
    if ( ( !was_used && *next_index == MAX_GOALS ) ||
	 i_e->endpt->is_goal > MAX_GOALS ) {
      printf( "\n\nipp: increase MAX_GOALS( preset value: %d )", MAX_GOALS );
      exit( 1 );
    }
    (i_e->endpt->is_goal_for)[i_e->endpt->is_goal-1] = op;
    if ( !was_used ) goals_at[time-1][(*next_index)++] = i_e->endpt;
  }

  /* ... and the effect conditions 
   *
   * NOTE: only put goals into the array that aren't already there
   */
  for ( i_e = conditions; i_e; i_e = i_e->next ) {
    was_used = i_e->endpt->is_goal++;
    if ( ( !was_used && *next_index == MAX_GOALS ) ||
	 i_e->endpt->is_goal > MAX_GOALS ) {
      printf( "\n\nipp: increase MAX_GOALS( preset value: %d )", MAX_GOALS ); 
      exit( 1 );
    }
    (i_e->endpt->is_goal_for)[i_e->endpt->is_goal-1] = op;
    if ( !was_used ) goals_at[time-1][(*next_index)++] = i_e->endpt;
  }

  /* now mark the facts that will be true for sure as being so
   * ( used for speedup ( skip over goals ) and in minimality check )
   */
  for ( i_ce = op->add_edges; i_ce; i_ce = i_ce->next )
    if ( !i_ce->conditions ) i_ce->endpt->is_true++;
  
  /* and say we managed */
  return TRUE;
  
}


/* this one simply puts an op out of the plan, i.e.
 * undoes the information in the graph and the global arrays
 */
void untry_op( vertex_list op, edge_list conditions,
               int *next_index, int time )

{

  edge_list i_e;
  cond_edge_list i_ce;
  int was_used;

  /* undo is true info */
  for ( i_ce = op->add_edges; i_ce; i_ce = i_ce->next )
    if ( !i_ce->conditions ) i_ce->endpt->is_true--;

  /* put the effect conditions out of the new-goal array */
  for ( i_e = conditions; i_e; i_e = i_e->next ) {
    was_used = --i_e->endpt->is_goal;
    (i_e->endpt->is_goal_for)[i_e->endpt->is_goal] = NULL;
    if ( !was_used ) goals_at[time-1][--(*next_index)] = NULL;
  }

  /* same for the preconds */
  for ( i_e = op->precond_edges; i_e; i_e = i_e->next ) {
    was_used = --i_e->endpt->is_goal;
    (i_e->endpt->is_goal_for)[i_e->endpt->is_goal] = NULL;
    if ( !was_used ) goals_at[time-1][--(*next_index)] = NULL;
  }

  /* and, finally, remove the op from the plan */
  was_used = --op->is_used;
  if( !was_used ) ops_at[time-1][--num_ops_at[time-1]] = NULL;

}


/* routine for building up SS, the 'Set of Set'
 *
 * look for bad conditional effects of the used ops 
 * and put their condition lists( without the new goals )
 * into the return list of edge list
 *
 * NOTE: SS can be non-empty, we just add new lists to those
 *       which are already there.
 */
list_of_edge_list compute_SS( int time, list_of_edge_list SS )

{

  list_of_edge_list result = SS, temp;/* used for building up result list */
  cond_edge_list i_ce;/* index */
  int i, j;/* inidizes */
  vertex_list op, ft;/* helpers */

  /* for all ops at time step time-1 do */
  for ( i=0; i<num_ops_at[time-1]; i++ ) {
    op = ops_at[time-1][i];
    /* for all conditional del effects */
    for ( i_ce = op->del_edges; i_ce; i_ce = i_ce->next ) {
      if ( i_ce->conditions ) {
        ft = i_ce->endpt;
        /* if goal is deleted */
        if ( ft->is_goal ||
           ( ft->prev_time && ft->prev_time->is_goal ) ) {

          /* if it is a new goal: see if it's the goal of someone else */     
          if ( !ft->is_goal ) {
            ft = ft->prev_time;
            for ( j=0; j<ft->is_goal; j++ )
              if ( (ft->is_goal_for)[j] != op ) break;
            if ( j == ft->is_goal ) continue;
	  }

          temp = ( list_of_edge_list )
                 calloc( 1, sizeof( list_of_edge_list_elt ) );
	  CHECK_MEMORY(temp);
          temp->list = without_new_goals( i_ce->conditions );
          /* store the op for the effect; need this info later in
           * complete_critics_... for knowing if the op adds it's own
           * critic
           */
          temp->op = ops_at[time-1][i];
          temp->next = result;
          result = temp;
        }
      }
    }
    /* for all conditional add effects */
    for ( i_ce = ops_at[time-1][i]->add_edges; i_ce; i_ce = i_ce->next ) {
      if ( i_ce->conditions ) {
        if ( i_ce->endpt->is_critic ) {
          temp = ( list_of_edge_list )
                 calloc( 1, sizeof( list_of_edge_list_elt ) );
	  CHECK_MEMORY(temp);
          temp->list = without_new_goals( i_ce->conditions );
          temp->next = result;
          temp->op = ops_at[time-1][i];
          result = temp;
        }
      }
    }
  }

  return result;

}


/* this one chooses a 'hitting set' out of SS, i.e. one fact from
 * each edge list; is called in search as a for - next loop:
 * first call with 3rd argument = INIT to initialise the choice,
 * then continue calling with 3rd argument = EXEC until
 * the return value is FALSE.
 */
BOOLEAN choose_critics( int time, list_of_edge_list SS, BOOLEAN initialise )

{

  static BOOLEAN first_call = TRUE;/* flag */
  static list_of_edge_list *ss;/* array for lists */
  static index_list_of_edge_list *index;/* array for indexs */

  list_of_edge_list i_list;/* index */
  index_list_of_edge_list i_index;/* index */
  int i,n;/* help for counting position is ss[time-1] */
  vertex_list ft;

  /* get memory for the arrays on first call */
  if ( first_call ) {
    ss = ( list_of_edge_list * )
         calloc( MAX_PLAN, sizeof( list_of_edge_list ) );
    CHECK_MEMORY(ss);
    index = ( index_list_of_edge_list * )
            calloc( MAX_PLAN, sizeof( index_list_of_edge_list ));
    CHECK_MEMORY(index);
    first_call = FALSE;
  }

  /* is it the first call on this SS ? */
  if ( initialise ) {

    /* first look if one of the lists is empty
     * ( happens if all effect conditions are contained in the new goals )
     *
     * if so, can't prevent this effect->can't choose any hitting
     * critic set -> return FALSE at once
     */
    for( i_list = SS; i_list; i_list = i_list->next )
      if ( !i_list->list ) return FALSE;

    /* setup the SS info
     *
     * NOTE: the function is called on mixed up time steps
     *       ( due to search recursion pattern ),
     *       so the statics have to be arrays of length MAX_PLAN
     */
    ss[time-1] = SS;
    /* this adds an 'i am the <n>th list information to each
     * edgelist, which we need later
     */
    free(index[time-1]);  /* Jana to eliminate memory leak */
    index[time-1] = make_index( ss[time-1], 0 );
  } else {

    /* first undo the critic info in the graph
     *
     * NOTE: the first fact in each list of index[time-1] is
     *       the critic we're currently using
     */ 
    for ( i_index = index[time-1]; i_index; i_index = i_index->next ) {
      ft = i_index->list->endpt;
      (ft->is_critic_for)[--ft->is_critic] = NULL;
    }

    /* now do some sort of 'increment counter' algorithm:
     * for all the indizes do:
     *   put index on the next critic in the list
     *   if there is no more critic there
     *     put it back on the first critic in the list
     *   else
     *     break
     * endfor
     * once we put all the indizes back to the start we know
     * that we tried all hitting sets and return FALSE.
     */
    for ( i_index = index[time-1]; i_index; i_index = i_index->next ) {
      i_index->list = i_index->list->next;
      if ( !i_index->list ) {
        /* that's the bit where we need the '<n>th list' info:
         * find the right list for our index in the stored ss[time-1] 
         * info and set index to the start of it.
         */
        n = 0;
        i_list = ss[time-1];
        while ( n++ < i_index->number ) i_list = i_list->next;
        i_index->list = i_list->list;
      } else {
        break;
      }
    }
    if ( !i_index ) return FALSE;
  }

  /* now the indizes are pointing on our new critic set;
   * add graph info and build up global array
   */
  for ( i=0; i<num_critics_at[time-1]; i++ )
    critics_at[time-1][i]->is_critic=0;
  num_critics_at[time-1] = 0;
  for ( i_index = index[time-1]; i_index; i_index = i_index->next ) {
    ft = i_index->list->endpt;
    /* say for which op it is critic; need this in complete_critics_... */
    (ft->is_critic_for)[ft->is_critic++] = i_index->op;
    if ( ft->is_critic == 1 ) {
      if ( num_critics_at[time-1] == MAX_GOALS ) {
	printf( "\n\nipp: increase MAX_GOALS( preset value: %d )",MAX_GOALS );
	exit( 1 );
      }
      critics_at[time-1][num_critics_at[time-1]++] = ft;
    }
  }

  /* say that we found a valid hitting set */
  return TRUE;

}
