/* (C) Copyright 1997 Albert Ludwigs University Freiburg
 *     Institute of Computer Science
 *
 * All rights reserved. Use of this software is permitted for 
 * non-commercial research purposes, and it may be copied only 
 * for that use.  All copies must include this copyright message.
 * This software is made available AS IS, and neither the authors
 * nor the  Albert Ludwigs University Freiburg make any warranty
 * about the software or its performance. 
 */

#ifndef lint
static char rcsid[] = "$Id: util_search.c,v 1.5 1998/03/04 13:28:01 ipp Exp $";
#endif /* lint */

/*
 * some helpers for searching
 */

#include"ipp.h"/* defines, data structures, fn prototypes, global variables */

/*
 * as it's name says, it does some setup utilities
 * for initialising the search process
 */
int setup( int max_time )

{

  int n = 0;/* number of goals at max time */
  vertex_list ft;/* helper */
  token_list i_t;/* helper */

  /* say that the facts in initial level are true already */
  get_next( fact_table[0], INIT );
  while ( ( ft = get_next( fact_table[0], EXEC ) ) != NULL ) ft->is_true = 1;

  /* make more space for global arrays of arrays */

  /* this stores goals at each time step, number is a parameter of search */
  if ( goals_at != NULL ) free( goals_at );
  goals_at = ( goal_array * ) calloc( max_time + 1, sizeof( goal_array ) );
  CHECK_MEMORY(goals_at);

  /* this stores the ops */
  if ( ops_at != NULL ) free( ops_at );
  ops_at = ( goal_array * ) calloc( max_time, sizeof( goal_array ) );
  CHECK_MEMORY(ops_at);
  /* also need to know how many there are */
  if ( num_ops_at != NULL ) free( num_ops_at );
  num_ops_at = ( int * ) calloc( max_time, sizeof( int ) );
  CHECK_MEMORY(num_ops_at);

  /* this stores the critics */
  if ( critics_at != NULL ) free( critics_at );
  critics_at = ( goal_array * ) calloc( max_time + 1, sizeof( goal_array ) );
  CHECK_MEMORY(critics_at);
  /* and their number */
  if ( num_critics_at != NULL ) free( num_critics_at );
  num_critics_at = ( int * ) calloc( max_time + 1, sizeof( int ) ); 
  CHECK_MEMORY(num_critics_at);

  /* setup the goal array at max time: lookup global 
   * goal strings from fact_table[max time]
   */
  for ( i_t = goal_facts; i_t; i_t = i_t->next ) {
    ft = lookup_from_table( fact_table[max_time], i_t->item );
    goals_at[max_time][n++] = ft;
    if ( n > MAX_GOALS ) {
      printf( "\n\nipp: increase MAX_GOALS( preset value: %d )", MAX_GOALS );
      exit( 1 );
    }
  }
  /* and the critic array at max time, which is empty */
  num_critics_at[max_time] = 0;

  /* return value is number of goals to be achieved */
  return n;

}


/* says if the successor of a fact ft is unconditionally deleted
 * by a chosen op at time - 1 other than op_
 *
 * ( this is used in choosing ops: see if an effect condition is
 *   being deleted by an already chosen parallel op )
 */
BOOLEAN is_deleted( vertex_list ft, int time, vertex_list op_ )

{

  cond_edge_list i_ce;
  int i;

  /* for all chosen ops at time - 1 do */
  for ( i=0; i<num_ops_at[time-1]; i++ ) {
    if( ops_at[time-1][i] == op_ ) continue;
    /* if op != op_ 
     * for all del edges of op do
     */
    for ( i_ce = ops_at[time-1][i]->del_edges; i_ce; i_ce = i_ce->next )
      /* if it unconditionally deletes the ft at next time,
       * return TRUE
       */
      if ( !i_ce->conditions && 
           i_ce->endpt->prev_time == ft ) return TRUE;
  }

  return FALSE;

}


/* says if a future goal has been cut off:
 * check all goals at time from 0 to max index n
 * if one of the ops making it true is still possible
 * ( given the choice of our so far used ops )
 * if there is no such op for one goal, we return false.
 */
BOOLEAN goals_still_possible( int n, int time, vertex_list op )

{

  int i;
  cond_edge_list i_ce;

  /* for all future goals do */
  for ( i=0; i<n; i++ ){
    /* for all add edges do */
    for ( i_ce = goals_at[time][i]->add_edges; i_ce; i_ce = i_ce->next )
      if ( !are_mutex( op, i_ce->endpt ) ) break;/* found a good one */
    if ( !i_ce ) return FALSE;
  }

  return TRUE;

} 


/* this makes up a new edge list, leaving out those vertexes
 * ( i.e. facts ) which are marked as being goals.
 *
 * used in compute_SS for leaving the new goals out of the
 * critic lists
 */
edge_list without_new_goals( edge_list l )

{

  edge_list result = NULL, temp, i_e;/* result, helpers */

  /* kein kommentar */
  for ( i_e = l; i_e; i_e = i_e->next ) {
    if ( !i_e->endpt->is_goal ) {
      temp = ( edge_list ) calloc(1, sizeof( edge_list_elt ) );
      CHECK_MEMORY(temp);
      temp->endpt = i_e->endpt;
      temp->next = result;
      result = temp;
    }
  }

  return result;

}


/* this is the minimality check:
 * see if there is an operator that we don't actually need.
 *
 * first increment the is true values of all facts that are
 * made true by a conditional effect of a used operator where
 * the effect conditions are completely contained in the new goals
 * ( the unconditional is true values are already set during 
 *   operator set setup in search )
 * then see for each op if there is a goal which he is the only
 * one to make true.
 * if we find an op that doesn't do so, return FALSE.
 */
BOOLEAN action_set_is_minimal( int time )

{

  int i;
  edge_list i_e;
  cond_edge_list i_ce;
  BOOLEAN result = TRUE;

  /* for all used ops do */
  for ( i=0; i<num_ops_at[time-1]; i++ ) 
    /* for all add effects do */
    for ( i_ce = ops_at[time-1][i]->add_edges; i_ce; i_ce = i_ce->next )
      /* if effect is conditional */
      if ( i_ce->conditions ) {
        /* for all conditions do */
        for ( i_e = i_ce->conditions; i_e; i_e = i_e->next )
          if ( !i_e->endpt->is_goal ) break;/* found a bad one */
        if ( !i_e ) i_ce->endpt->is_true++;
      }

  /* for all used ops do */
  for ( i=0; i<num_ops_at[time-1]; i++ ) {
    /* for all add effects do */
    for ( i_ce = ops_at[time-1][i]->add_edges; i_ce; i_ce = i_ce->next ) {
      if ( !i_ce->endpt->is_goal ) continue;/* not a goal */
      if ( i_ce->endpt->is_true == 1 ) break;/* good op */
    }
    if( !i_ce ) {/* didn't need that one */
      result = FALSE;
      break;
    }
  }

  /* undo the is true changings that we made cause we need them
   * in the search recursion pattern
   */
  for ( i=0; i<num_ops_at[time-1]; i++ ) 
    for ( i_ce = ops_at[time-1][i]->add_edges; i_ce; i_ce = i_ce->next )
      if ( i_ce->conditions ) {
        for ( i_e = i_ce->conditions; i_e; i_e = i_e->next )
          if ( !i_e->endpt->is_goal ) break;
        if ( !i_e ) i_ce->endpt->is_true--;
      }

  /* and return the result */
  return result;

}


/* this is the routine that prints the actual output
 * of the whole thing to stdout
 */
void print_plan( int time ) 

{

  vertex_list op;/* helper */

  int i, j;/* index for time */
  BOOLEAN first;/* for better printing: is it the first one this level? */

  printf( "\nipp: found plan as follows\n\n" );

  /* for all non empty operator levels */
  for ( i=0; i<time; i++ ) {
    printf("time step %3d: ", i);
    first = TRUE;
    /* for all operators in plan */
    for ( j=0; j<num_ops_at[i]; j++ ) {
      op = ops_at[i][j];
      if ( op->is_noop ) continue;/* don't print noops */
      if ( first ) {
	printf("%s\n", op->name );
	first = FALSE;
      } else {
	printf( "               %s\n", op->name );
      }
    }
  }

  printf( "\n" );

}


/* used for making up choice in choose_critics;
 *
 * copys SS and adds an 'i am the <n>th list' information,
 * technical for the choose critics algorithm
 */
index_list_of_edge_list make_index( list_of_edge_list l, int n )

{

  index_list_of_edge_list result;

  if ( !l ) return NULL;

  result = ( index_list_of_edge_list )
           calloc( 1, sizeof( index_list_of_edge_list_elt ) );
  CHECK_MEMORY(result);
  result->list = l->list;
  result->number = n;
  result->op = l->op;
  result->next = make_index( l->next, n + 1 );

  return result;

}


BOOLEAN each_list_contains_dummy( list_of_edge_list SS )

{

  list_of_edge_list i;
  edge_list j;

  for ( i = SS; i; i = i->next ) {
    for ( j = i->list; j; j = j->next )
      if ( j->endpt->is_dummy ) break;
    if ( !j ) return FALSE;
  }
 
  return TRUE;

}


BOOLEAN cant_do_op( vertex_list op, int time )

{

  int i;

  if ( op->is_used ) return FALSE;

  for ( i=0; i<num_ops_at[time]; i++ ) {
    if ( are_mutex( ops_at[time][i], op ) ) return TRUE;
  }

  return FALSE;

}


BOOLEAN cant_do_ft( vertex_list ft, int time )

{

  int i;

  if ( ft->is_used ) return FALSE;

  for ( i=0; ( goals_at[time][i] != NULL ); i++ ) {
    if ( are_mutex( goals_at[time][i], ft ) ) return TRUE;
  }

  return FALSE;

}
